#!/bin/env python3
import os

from pwn import *
import vagd.virts.pwngd
from vagd import Vagd, Qegd, Shgd, Dogd, Logd, wrapper, Box

GDB_OFF = 0x555555554000
IP = ""
PORT = 0
BINARY = "./bin/sysinfo_stat"
ARGS = []
ENV = {}
API = True
GDB = """
b main
c"""

context.binary = exe = ELF(BINARY, checksec=False)
context.aslr = False

byt = lambda x: str(x).encode()

LOCKFILE = vagd.virts.pwngd.Pwngd.LOCKFILE


def test_lockfile(expected):
  with open(LOCKFILE) as lockfile:
    assert lockfile.read() == expected, "bad lockfile"


def stage(msg, *args):
  log.info("======== " + msg + " ========", *args)


def virts():
  stage("Testing Local")
  yield Logd(exe.path)

  stage("Testing Logging")
  context.log_level = "error"
  yield Logd(exe.path)
  context.log_level = "info"

  if args.VAGRANT:
    stage("Testing Vagrant")

    if os.path.exists(Vagd.VAGRANTFILE_PATH):
      os.system(f"VAGRANT_CWD={Vagd.LOCAL_DIR} vagrant destroy -f")
      os.remove(Vagd.VAGRANTFILE_PATH)

    vm = Vagd(
      exe.path,
      vbox=Box.VAGRANT_JAMMY64,
      packages=["cowsay"],
      tmp=True,
      fast=True,
      ex=True,
    )
    assert vm.is_new, "vm should be new"
    assert vm._ssh.which("cowsay"), "cowsay wasn't installed"
    test_lockfile(Vagd.TYPE)
    yield vm
    vm._ssh.close()

    stage("Testing Vagrant restore")
    vm = Vagd(exe.path, vbox=Box.VAGRANT_JAMMY64, tmp=True, fast=True, ex=True)
    assert not vm.is_new, "vm shouldn't be new, restored"
    yield vm
    vm._ssh.close()

    stage("Testing Vagrant restart")
    os.system(f"VAGRANT_CWD={Vagd.LOCAL_DIR} vagrant halt")
    vm = Vagd(exe.path, vbox=Box.VAGRANT_JAMMY64, tmp=True, fast=True, ex=True)
    assert not vm.is_new, "vm shouldn't be new, restarted"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")

  if not args.NODOGD:
    if os.path.exists(Dogd.LOCKFILE):
      os.remove(Dogd.LOCKFILE)
    stage("Testing Docker for Ubuntu")
    vm = Dogd(
      exe.path,
      image=Box.DOCKER_UBUNTU,
      packages=["cowsay"],
      tmp=True,
      ex=True,
      fast=True,
    )
    assert vm.is_new, "vm should be new"
    assert vm._ssh.which("cowsay"), "cowsay wasn't installed"
    yield vm
    vm._ssh.close()

    stage("Testing Docker for Ubuntu restore")
    vm = Dogd(exe.path, image=Box.DOCKER_UBUNTU, tmp=True, ex=True, fast=True)
    assert not vm.is_new, "vm shouldn't be new, restored"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")
    sleep(1)
    stage("Testing Docker for Ubuntu (root)")
    vm = Dogd(
      exe.path,
      user="root",
      image=Box.DOCKER_UBUNTU,
      tmp=True,
      ex=True,
      fast=True,
    )
    assert vm.is_new, "vm should be new"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")
    sleep(1)
    stage("Testing Docker for Alpine")
    vm = Dogd(
      exe.path,
      image=Box.DOCKER_ALPINE,
      tmp=True,
      ex=True,
      fast=True,
    )
    assert vm.is_new, "vm should be new"
    yield vm
    vm._ssh.close()

    stage("Testing Docker for Alpine restore")
    vm = Dogd(
      exe.path,
      image=Box.DOCKER_ALPINE,
      tmp=True,
      ex=True,
      fast=True,
    )
    assert not vm.is_new, "vm shouldn't be new, restored"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")
    sleep(1)
    stage("Testing Docker for Alpine (root)")
    vm = Dogd(
      exe.path,
      image=Box.DOCKER_ALPINE,
      user="root",
      tmp=True,
      ex=True,
      fast=True,
    )
    assert vm.is_new, "vm should be new"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")

  if not args.NO_QEMU:
    stage("Testing Qemu")
    vm = Qegd(
      exe.path,
      img=Box.QEMU_UBUNTU,
      tmp=True,
      packages=["cowsay"],
      ex=True,
      fast=True,
    )
    assert vm.is_new, "vm should be new"
    assert vm._ssh.which("cowsay"), "cowsay wasn't installed"
    yield vm
    vm._ssh.close()

    stage("Testing Qemu restore")
    vm = Qegd(exe.path, img=Box.QEMU_UBUNTU, tmp=True, ex=True, fast=True)
    assert not vm.is_new, "vm shouldn't be new, restored"
    yield vm
    vm._ssh.close()

    os.system("vagd clean")
    sleep(1)
    stage("Testing Qemu (root)")
    vm = Qegd(
      exe.path,
      img=Box.QEMU_UBUNTU,
      user="root",
      tmp=True,
      ex=True,
      fast=True,
      root=True,
    )
    assert vm.is_new, "vm should be new"
    yield vm
    vm._ssh.close()
    user = vm._user
    port = vm._port

    stage("Testing SSH")
    yield Shgd(
      exe.path,
      user=user,
      port=port,
      keyfile=vm._ssh.keyfile,
      tmp=True,
      ex=True,
      fast=True,
    )


for virt in virts():
  t = virt.start(argv=ARGS, env=ENV, gdbscript=GDB, api=API)

  sleep(1)
  if args.GDB:
    g = wrapper.GDB(t)
    g.execute('p "PWN"')
    g.execute("c")

  out = b"\n".join(t.recvlines(3))

  log.info(out.decode())
  t.close()
  os.system("tmux kill-pane")

os.system("vagd clean")
sleep(1)
assert not os.path.exists(LOCKFILE), "lockfile shouldn't exist"

stage("Everything executed without errors")



from typing import Iterable, Optional

import pwnlib.args
import pwnlib.tubes

from vagd import helper
from vagd.virts.pwngd import Pwngd


class Logd(Pwngd):
  """
  local execution of binary

  :param binary: binary to execute
  """

  _binary: str

  def _vm_setup(self) -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def _ssh_setup(self) -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def __init__(self, binary: str, **kwargs):
    """
    :param binary: binary to execute
    """
    self._binary = binary

  def _sync(self, file: str) -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def _mount(self, remote_dir: str, local_dir: str) -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def _mount_lib(self, remote_lib: str = "/usr/lib") -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def system(self, cmd: str) -> None:
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def _install_packages(self, packages: Iterable):
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def put(self, file: str, remote: Optional[str] = None):
    """
    NOT IMPLEMENTED
    """
    helper.error("NOT IMPLEMENTED")

  def debug(self, **kwargs) -> pwnlib.tubes.process.process:
    """
    run binary with gdb locally
    :param kwargs: pwntool arguments
    :rtype: pwnlib.tubes.process.process
    """
    return self.pwn_debug(**kwargs)

  def pwn_debug(self, argv: Optional[list[str]] = None, **kwargs) -> pwnlib.tubes.process.process:
    """
    run binary with gdb locally
    :param argv: comandline arguments for binary
    :param kwargs: pwntool arguments
    :rtype: pwnlib.tubes.process.process
    """
    return pwnlib.gdb.debug([self._binary] + argv, **kwargs)

  def process(self, argv: Optional[list[str]] = None, **kwargs) -> pwnlib.tubes.process.process:
    """
    run binary locally
    :param argv: comandline arguments for binary
    :param kwargs: pwntool parameters
    :return: pwntools process
    """
    return pwnlib.tubes.process.process([self._binary] + argv, **kwargs)

  def start(
    self,
    argv: Optional[list[str]] = None,
    gdbscript: str = "",
    api: bool = False,
    **kwargs,
  ) -> pwnlib.tubes.process.process:
    """
    start binary locally and return pwnlib.tubes.process.process
    :param argv: commandline arguments for binary
    :param gdbscript: GDB script for GDB
    :param api: if GDB API should be enabled (experimental)
    :param kwargs: pwntool parameters
    :return: pwntools process, if api=True tuple with gdb api
    """
    helper.warn("running locally, only limited functions are supported")
    if pwnlib.args.args.GDB:
      return self.pwn_debug(argv=argv, gdbscript=gdbscript, api=api, **kwargs)
    else:
      return self.process(argv=argv, **kwargs)



import time

import pwnlib.tubes.ssh

from vagd import helper
from vagd.virts.pwngd import Pwngd


class Shgd(Pwngd):
  """
  ssh interface for pwntools

  :param binary: binary to execute
  :param user: ssh user
  :param host: ssh hostname
  :param port: ssh port
  :param keyfile: ssh keyfile (default in .vagd)
  :param kwargs: parameters to pass through to super
  """

  DEFAULT_HOST = "localhost"
  DEFAULT_PORT = 22
  DEFAULT_USER = "root"

  _user: str
  _host: str
  _port: int
  _keyfile: str
  _ssh: pwnlib.tubes.ssh.ssh

  def bind(self, port: int) -> int:
    """
    bind port from ssh connection locally
    :param port:
    :return:
    """

    remote = self._ssh.connect_remote("127.0.0.1", port)
    listener = pwnlib.tubes.listen.listen(0)
    port = listener.lport

    # Disable showing GDB traffic when debugging verbosity is increased
    remote.level = "error"
    listener.level = "error"

    # Hook them up
    remote.connect_both(listener)

    return port

  def _vm_setup(self) -> None:
    """
    pass
    """
    pass

  _TRIES = 3  # three times the charm

  def _ssh_setup(self) -> None:
    """
    setup ssh connection
    """
    progress = helper.progress("connecting to ssh")
    for i in range(Shgd._TRIES):
      try:
        self._ssh = pwnlib.tubes.ssh.ssh(
          user=self._user,
          host=self._host,
          port=self._port,
          keyfile=self._keyfile,
          ignore_config=True,
        )
        progress.success("Done")
        break
      except Exception as e:
        if i + 1 == Shgd._TRIES:
          progress.failure("%s", e)
          helper.error("Failed to connect to ssh")
        else:
          progress.status("Trying again")
        time.sleep(1 if i == 0 else 10)

  def __init__(
    self,
    binary: str,
    user: str = DEFAULT_USER,
    host: str = DEFAULT_HOST,
    port: int = DEFAULT_PORT,
    keyfile: str = Pwngd.KEYFILE,
    **kwargs,
  ):
    """

    :param binary: binary to execute
    :param user: ssh user
    :param host: ssh hostname
    :param port: ssh port
    :param keyfile: ssh keyfile (default in .vagd)
    :param kwargs: parameters to pass through to super
    """
    self._user = user
    self._host = host
    self._port = port
    self._keyfile = keyfile

    self._ssh_setup()

    super().__init__(binary=binary, **kwargs)



from collections.abc import Callable, Sequence

import gdb

class ThreadEvent:
    inferior_thread: gdb.InferiorThread

class ContinueEvent(ThreadEvent): ...

class ContinueEventRegistry:
    def connect(self, __object: Callable[[ContinueEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[ContinueEvent], object]) -> None: ...

cont: ContinueEventRegistry

class ExitedEvent:
    exit_code: int
    inferior: gdb.Inferior

class ExitedEventRegistry:
    def connect(self, __object: Callable[[ExitedEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[ExitedEvent], object]) -> None: ...

exited: ExitedEventRegistry

class StopEvent(ThreadEvent):
    stop_signal: str

class BreakpointEvent(StopEvent):
    breakpoints: Sequence[gdb.Breakpoint]
    breakpoint: gdb.Breakpoint

class StopEventRegistry:
    def connect(self, __object: Callable[[StopEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[StopEvent], object]) -> None: ...

stop: StopEventRegistry

class NewObjFileEvent:
    new_objfile: gdb.Objfile

class NewObjFileEventRegistry:
    def connect(self, __object: Callable[[NewObjFileEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[NewObjFileEvent], object]) -> None: ...

new_objfile: NewObjFileEventRegistry

class ClearObjFilesEvent:
    progspace: gdb.Progspace

class ClearObjFilesEventRegistry:
    def connect(self, __object: Callable[[ClearObjFilesEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[ClearObjFilesEvent], object]) -> None: ...

clear_objfiles: ClearObjFilesEventRegistry

class InferiorCallEvent: ...

class InferiorCallPreEvent(InferiorCallEvent):
    ptid: gdb.InferiorThread
    address: gdb.Value

class InferiorCallPostEvent(InferiorCallEvent):
    ptid: gdb.InferiorThread
    address: gdb.Value

class InferiorCallEventRegistry:
    def connect(self, __object: Callable[[InferiorCallEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[InferiorCallEvent], object]) -> None: ...

inferior_call: InferiorCallEventRegistry

class MemoryChangedEvent:
    address: gdb.Value
    length: int

class MemoryChangedEventRegistry:
    def connect(self, __object: Callable[[MemoryChangedEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[MemoryChangedEvent], object]) -> None: ...

memory_changed: MemoryChangedEventRegistry

class RegisterChangedEvent:
    frame: gdb.Frame
    regnum: str

class RegisterChangedEventRegistry:
    def connect(self, __object: Callable[[RegisterChangedEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[RegisterChangedEvent], object]) -> None: ...

register_changed: RegisterChangedEventRegistry

class BreakpointEventRegistry:
    def connect(self, __object: Callable[[gdb.Breakpoint], object]) -> None: ...
    def disconnect(self, __object: Callable[[gdb.Breakpoint], object]) -> None: ...

breakpoint_created: BreakpointEventRegistry
breakpoint_modified: BreakpointEventRegistry
breakpoint_deleted: BreakpointEventRegistry

class BeforePromptEventRegistry:
    def connect(self, __object: Callable[[], object]) -> None: ...
    def disconnect(self, __object: Callable[[], object]) -> None: ...

before_prompt: BeforePromptEventRegistry

class NewInferiorEvent:
    inferior: gdb.Inferior

class NewInferiorEventRegistry:
    def connect(self, __object: Callable[[NewInferiorEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[NewInferiorEvent], object]) -> None: ...

new_inferior: NewInferiorEventRegistry

class InferiorDeletedEvent:
    inferior: gdb.Inferior

class InferiorDeletedEventRegistry:
    def connect(self, __object: Callable[[InferiorDeletedEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[InferiorDeletedEvent], object]) -> None: ...

inferior_deleted: InferiorDeletedEventRegistry

class NewThreadEvent(ThreadEvent): ...

class NewThreadEventRegistry:
    def connect(self, __object: Callable[[NewThreadEvent], object]) -> None: ...
    def disconnect(self, __object: Callable[[NewThreadEvent], object]) -> None: ...

new_thread: NewThreadEventRegistry



import fileinput
import os
import re
from shutil import which
from typing import List, Optional

from vagd import helper, templates
from vagd.box import Box
from vagd.virts.pwngd import Pwngd
from vagd.virts.shgd import Shgd


class Vagd(Shgd):
  """
  | Vagrant Virtualization for pwntools

  :param binary: binary for VM debugging
  :param vbox: vagrant box to use
  :param vagrantfile: location of Vagrantfile
  :param packages: packages to install on vm
  :param kwargs: arguments to pass through to super

  | SSH from cmd:

  .. code-block::  bash

    vagd ssh
    # or
    VAGRANT_CWD=.vagd vagrant ssh

  | halt from cmd

  .. code-block:: bash

    VAGRANT_CWD=.vagd vagrant halt

  | destroy from cmd

  .. code-block:: bash

    vagd clean
    # or
    VAGRANT_CWD=.vagd vagrant destroy
  """

  VAGRANTFILE_PATH = Pwngd.LOCAL_DIR + "Vagrantfile"
  VAGRANTFILE_BOX = "config.vm.box"
  VAGRANT_BOX = Box.VAGRANT_JAMMY64
  KEYFILE = Pwngd.LOCAL_DIR + ".vagd/.vagrant/machines/default/virtualbox/private_key"
  TYPE = "vagd"

  _box: str
  _vagrantfile: str
  _v: object

  def _get_box(self) -> str:
    """
    returns box of current vagrantfile
    @:rtype box name of Vagrantfile
    """
    with open(self._vagrantfile, "r") as vagrantfile:
      for line in vagrantfile.readlines():
        if Vagd.VAGRANTFILE_BOX in line:
          pattern = rf'{Vagd.VAGRANTFILE_BOX} = "(.*?)"'
          match = re.search(pattern, line, re.DOTALL)
          if match:
            return match.group(1)
    return ""

  def _vm_setup(self) -> None:
    """
    setup vagrant machine creates new one if no Vagrantfile is specified or box does not match
    """

    if self._get_box() != self._box:
      helper.info("new box detected destroying old machine")
      self._v.destroy()
      for line in fileinput.input(self._vagrantfile, inplace=True):
        if Vagd.VAGRANTFILE_BOX in line:
          line = f'{Vagd.VAGRANTFILE_BOX} = "{self._box}"\n'
        print(line, end="")

    if self._v.status()[0].state == "not_created":
      self.is_new = True
      helper.info("initialing new vagrant vm might take a while")
      self._v.up()

    if self._v.status()[0].state != "running":
      helper.info("starting existing vagrant machine")
      self._v.up()

  def __init__(
    self,
    binary: str,
    vagrantfile: str = VAGRANTFILE_PATH,
    vbox: Optional[str] = None,
    packages: Optional[List[str]] = None,
    **kwargs,
  ):
    """

    :param binary: binary for VM debugging
    :param vbox: vagrant box to use
    :param vagrantfile: location of Vagrantfile
    :param packages: packages to install on vm
    :param kwargs: arguments to pass through to super
    """
    import vagrant

    helper.warn("The 'Vagd' object is deprecated, use 'Qegd' instead")

    if packages is None:
      packages = list()

    if not which("vagrant"):
      helper.error("vagrant isn't installed")

    if not os.path.exists(Pwngd.LOCAL_DIR):
      os.makedirs(Pwngd.LOCAL_DIR)

    self._vagrantfile = vagrantfile

    if vbox is None:
      if os.path.exists(vagrantfile):
        vbox = self._get_box()
      else:
        vbox = Vagd.VAGRANT_BOX

    self._box = vbox

    self._lock(Vagd.TYPE)
    if not os.path.isfile(self._vagrantfile):
      helper.info("creating new Vagrantfile")
      vagrant_config = templates.VAGRANT_TEMPLATE.format(box=self._box)
      with open(self._vagrantfile, "w") as file:
        file.write(vagrant_config)

    self._v = vagrant.Vagrant(os.path.dirname(self._vagrantfile))

    self._vm_setup()

    packages += Pwngd.DEFAULT_PACKAGES

    super().__init__(
      binary=binary,
      user=self._v.user(),
      host=self._v.hostname(),
      port=int(self._v.port()),
      packages=packages,
      keyfile=self._v.keyfile(),
      **kwargs,
    )



import os
from abc import ABC, abstractmethod
from shutil import which
from typing import Iterable, List, Union, Optional

import pwnlib.args
import pwnlib.filesystem
import pwnlib.gdb
import pwnlib.tubes

from vagd import helper


class Pwngd(ABC):
  """
  start binary on remote and return pwnlib.tubes.process.process

  :param binary: binary for VM debugging
  :param libs: download libraries (using ldd) from VM
  :param files: other files or directories that need to be uploaded to VM
  :param packages: packages to install on vm
  :param symbols: additionally install libc6 debug symbols
  :param tmp: if a temporary directory should be created for files
  :param gdbsrvport: specify static gdbserver port, REQURIES port forwarding to localhost
  :param fast: mounts libs locally for faster symbol extraction (experimental)
  :param ex: if experimental features should be enabled
  """

  LOCAL_DIR = "./.vagd/"
  HOME_DIR = os.path.expanduser("~/.local/share/vagd/")
  SYSROOT = LOCAL_DIR + "sysroot/"
  SYSROOT_LIB_DEBUG = SYSROOT + "lib/debug"
  LOCKFILE = LOCAL_DIR + "vagd.lock"
  KEYFILE = HOME_DIR + "keyfile"
  PUBKEYFILE = KEYFILE + ".pub"
  DEFAULT_PORT = 2222
  STATIC_GDBSRV_PORT = 42069

  is_new: bool = False
  _path: str
  _gdbsrvport: int
  _binary: str
  _ssh: pwnlib.tubes.ssh.ssh
  _experimental: bool
  _fast: bool

  def __init__(
    self,
    binary: str,
    libs: bool = False,
    files: Optional[Union[str, list[str]]] = None,
    packages: Optional[List[str]] = None,
    symbols: bool = True,
    tmp: bool = False,
    gdbsrvport: int = 0,
    root: bool = False,
    fast: bool = False,
    ex: bool = False,
  ):
    self._path = binary
    self._gdbsrvport = gdbsrvport
    self._binary = "./" + os.path.basename(binary)

    pwnlib.context.context.ssh_session = self._ssh

    if tmp:
      self._ssh.set_working_directory()

    if self._sync(self._path):
      self.system("chmod +x " + self._binary)

    if self.is_new and libs:
      if not (os.path.exists(Pwngd.LIBS_DIRECTORY)):
        os.makedirs(Pwngd.LIBS_DIRECTORY)

      self.libs(Pwngd.LIBS_DIRECTORY)

    if self.is_new and packages is not None:
      if symbols:
        packages.append(Pwngd.LIBC6_DEBUG)
      try:
        elf = pwnlib.elf.ELF(binary, checksec=False)
        if elf.arch == "i386":
          packages.append(Pwngd.LIBC6_I386)
      except:
        helper.warn("failed to get architecture from binary")
      self._install_packages(packages)

    self._fast = fast
    self._experimental = ex

    if self._fast:
      if self._experimental:
        self._mount_root()
      else:
        helper.error("requires experimental features, activate with ex=True")

    # Copy files to remote
    if isinstance(files, str):
      self._sync(files)
    elif hasattr(files, "__iter__"):
      for file in files:
        self._sync(file)

  @abstractmethod
  def _vm_setup(self) -> None:
    """
    setup vagrant machine creates new one if no Vagrantfile is specified or box does not match
    """
    pass

  @abstractmethod
  def _ssh_setup(self) -> None:
    """
    setup ssh connection
    """
    pass

  def _sync(self, file: str) -> bool:
    """
    upload file on remote if not exist
    :type file: file to upload
    :return: if the file was uploaded
    """
    sshpath = pwnlib.filesystem.SSHPath(file)
    if not sshpath.exists():
      self.put(file)
      return True
    return False

  _SSHFS_TEMPLATE = "sshfs -p {port} -o StrictHostKeyChecking=no,ro,IdentityFile={keyfile} {user}@{host}:{remote_dir} {local_dir}"

  def _mount(self, remote_dir: str, local_dir: str) -> None:
    """
    mount remote dir on local wiith sshfs
    :param remote_dir: directory on remote to mount
    :param local_dir: local mount point
    """
    if not which("sshfs"):
      helper.error("sshfs isn't installed")
    cmd = Pwngd._SSHFS_TEMPLATE.format(
      port=self._ssh.port,
      keyfile=self._ssh.keyfile,
      user=self._ssh.user,
      host=self._ssh.host,
      remote_dir=remote_dir,
      local_dir=local_dir,
    )
    helper.info(cmd)
    os.system(cmd)

  def _lock(self, typ: str):
    if not os.path.exists(Pwngd.LOCAL_DIR):
      os.makedirs(Pwngd.LOCAL_DIR)

    with open(Pwngd.LOCKFILE, "w") as lfile:
      lfile.write(typ)

  def _mount_root(self, remote_lib: str = "/") -> None:
    """
    mount the lib directory of remote
    """
    if not os.path.exists(Pwngd.SYSROOT):
      os.makedirs(Pwngd.SYSROOT)
    if not os.path.ismount(Pwngd.SYSROOT):
      helper.info("mounting libs in sysroot")
      self._mount(remote_lib, Pwngd.SYSROOT)

  def system(self, cmd: str) -> pwnlib.tubes.ssh.ssh_channel:
    """
    executes command on vm, interface to  pwnlib.tubes.ssh.ssh.system

    :param cmd: command to execute on vm
    :return: returns
    """
    return self._ssh.system(cmd)

  DEFAULT_PACKAGES = ["gdbserver", "python3", "sudo"]
  LIBC6_DEBUG = "libc6-dbg"
  LIBC6_I386 = "libc6-i386"

  def _install_packages(self, packages: Iterable[str]):
    """
    install packages on remote machine

    :param packages: packages to install on remote machine
    """
    helper.info(f"installing packages: {' '.join(packages)}")
    self.system("sudo apt update").recvall()
    packages_str = " ".join(packages)
    self.system(f"sudo DEBIAN_FRONTEND=noninteractive apt install -y {packages_str}").recvall()

  def put(self, file: str, remote: Optional[str] = None):
    """
    upload file or dir on vm,

    :param file: file to upload
    :param remote: remote location of file, working directory if not specified
    :return: returns
    """
    if os.path.isdir(file):
      self._ssh.upload_dir(file, remote=remote)
    else:
      self._ssh.upload(file, remote=remote)

  def pull(self, file: str, local: Optional[str] = None):
    """
    download file or dir on vm,

    :param file: remote location of file, working directory if not specified
    :param local: local location of file, current directory if not specified
    :return: returns
    """
    sshpath = pwnlib.filesystem.SSHPath(os.path.basename(file))
    if sshpath.is_dir():
      self._ssh.download_dir(file, local=local)
    else:
      self._ssh.download_file(file, local=local)

  LIBS_DIRECTORY = "libs"

  def libs(self, directory: str):
    """
    Downloads the libraries referred to by a file.
    This is done by running ldd on the remote server, parsing the output and downloading the relevant files.

    directory(str): Output directory
    :return:
    """
    for lib in self._ssh._libs_remote(self._binary).keys():
      self.pull(lib, directory + "/" + os.path.basename(lib))

  def debug(
    self,
    argv: Optional[list[str]] = None,
    gdb_args: Optional[list[str]] = None,
    gdbscript: str = "",
    sysroot: Optional[str] = None,
    sysroot_debug: Optional[str] = None,
    **kwargs,
  ) -> pwnlib.tubes.ssh.ssh_channel:
    """
    run binary in vm with gdb (pwnlib feature set)

    :param argv: comandline arguments for binary
    :param gdb_args: gdb args to forward to gdb
    :param gdbscript: GDB script for GDB
    :param sysroot: sysroot dir
    :param sysroot_debug: sysroot debug lib dir
    :param kwargs: pwntool parameters
    :return: pwntools process
    """
    if argv is None:
      argv = list()

    if gdb_args is None:
      gdb_args = list()

    if self._fast:
      if sysroot is not None:
        helper.warn("fast enabled but sysroot set, sysroot is ignored")
      sysroot = Pwngd.SYSROOT

    if sysroot_debug is not None and sysroot is None:
      helper.warn("sysroot_debug set, but sysroot isn't ignored")

    if sysroot is not None:
      if sysroot_debug == None:
        sysroot_debug = Pwngd.SYSROOT_LIB_DEBUG
      gdbscript = f"set debug-file-directory {sysroot_debug}\n" + gdbscript

    gdb_args += ["-ex", f"file -readnow {self._path}"]

    return pwnlib.gdb.debug(
      [self._binary] + argv,
      ssh=self._ssh,
      gdb_args=gdb_args,
      port=self._gdbsrvport,
      gdbscript=gdbscript,
      sysroot=sysroot,
      **kwargs,
    )

  def process(self, argv: Optional[list[str]] = None, **kwargs) -> pwnlib.tubes.ssh.ssh_channel:
    """
    run binary in vm as process

    :param argv: comandline arguments for binary
    :param kwargs: pwntool parameters
    :return: pwntools process
    """
    if argv is None:
      argv = list()
    return self._ssh.process([self._binary] + argv, **kwargs)

  def start(
    self,
    argv: list[str] = None,
    gdbscript: str = "",
    api: bool = False,
    sysroot: Optional[str] = None,
    sysroot_debug: Optional[str] = None,
    gdb_args: Optional[list[str]] = None,
    **kwargs,
  ) -> pwnlib.tubes.ssh.ssh_channel:
    """
    start binary on remote and return pwnlib.tubes.process.process

    :param argv: commandline arguments for binary
    :param gdbscript: GDB script for GDB
    :param api: if GDB API should be enabled
    :param sysroot: sysroot dir
    :param sysroot_debug: sysroot debug lib dir
    :param gdb_args: extra gdb args
    :param kwargs: pwntool parameters
    :return: pwntools process, if api=True tuple with gdb api
    """
    if pwnlib.args.args.GDB:
      return self.debug(
        argv=argv,
        gdbscript=gdbscript,
        gdb_args=gdb_args,
        sysroot=sysroot,
        api=api,
        **kwargs,
      )
    else:
      return self.process(argv=argv, **kwargs)



import os
import time
from shutil import copyfile, which
from typing import Dict, List, Optional
from urllib.parse import urlparse
from urllib.request import urlretrieve

import pwnlib

from vagd import helper
from vagd.box import Box
from vagd.virts.pwngd import Pwngd
from vagd.virts.shgd import Shgd


class Qegd(Shgd):
  """
  | QEMU Virtualization for pwntools

  :param binary: binary for VM debugging
  :param img: qemu image to use (requires ssh)
  :param user: user inside qemu image
  :param ports: forwarded ports
  :param packages: packages to install on vm
  :param arm: emulate arm in qemu
  :param qemu: qemu cmd
  :param cpu: value for :code -cpu
  :param memory: value for :code -m
  :param cores: value for :code -smp
  :param machine: value for :code -machine
  :param bios: value for :code -bios
  :param custom: custom qemu arguments
  :param detach: run qemu in new terminal
  :param kwargs: parameters to pass through to super

  | SSH from cmd

  .. code-block:: bash

      vagd ssh
      # or
      ssh -o "StrictHostKeyChecking=no" -i ~/.share/local/vagd/keyfile -p $(cat .vagd/qemu.lock) ubuntu@0.0.0.0

  | Kill from cmd:

  .. code-block:: bash

      vagd clean
      # or
      kill $(pgrep qemu)

  | Qemu images are cached in the home directory: :code:`~/.share/local/vagd/qemu-imgs/`
  |
  | current used images are stored in the local directory: :code:`./.vagd/current.img`
  | These should be deleted automatically, but if a machine gets improperly stopped
  | (shutdown host while vm is running) it might remain and use up space. You can find remaining images with:

  .. code-block:: bash

      find ~/ -name current.img
      rm <path/current.img>

  """

  DEFAULT_IMG = Box.QEMU_NOBLE
  QEMU_DIR = Pwngd.LOCAL_DIR
  IMGS_DIR = Pwngd.HOME_DIR + "qemu-imgs/"
  DEFAULT_USER = "vagd"
  DEFAULT_HOST = "0.0.0.0"
  TYPE = "qegd"
  DEFAULT_PORT = 2222

  DEFAULT_QEMU_CMD = "qemu-system-x86_64"
  DEFAULT_QEMU_ARM_CMD = "qemu-system-aarch64"

  DEFAULT_QEMU_MACHINE_PREFIX = "-machine"
  DEFAULT_QEMU_MACHINE = "accel=kvm,type=q35"
  DEFAULT_QEMU_ARM_MACHINE = "virt"

  DEFAULT_QEMU_CPU_PREFIX = "-cpu"
  DEFAULT_QEMU_CPU = "host"
  DEFAULT_QEMU_ARM_CPU = "cortex-a72"

  DEFAULT_QEMU_CORES_PREFIX = "-smp"
  DEFAULT_QEMU_CORES = "2"

  DEFAULT_QEMU_BIOS_PREFIX = "-bios"
  DEFAULT_QEMU_ARM_BIOS = "/usr/share/edk2/aarch64/QEMU_EFI.fd"

  DEFAULT_QEMU_MEMORY_PREFIX = "-m"
  DEFAULT_QEMU_MEMORY = "2G"

  _img: str
  _local_img: str
  _user: str
  _host: str
  _port: int
  _forward: Dict[str, int]
  _qemu: str
  _cpu: str
  _cores: str
  _memory: str
  _bios: str
  _cores: str
  _machine: str
  _detach: bool

  def __init__(
    self,
    binary: str,
    img: str = DEFAULT_IMG,
    user: str = DEFAULT_USER,
    forward: Optional[Dict[str, int]] = None,
    packages: Optional[List[str]] = None,
    arm: bool = False,
    qemu: str = DEFAULT_QEMU_CMD,
    cpu: str = DEFAULT_QEMU_CPU,
    memory: str = DEFAULT_QEMU_MEMORY,
    machine: str = DEFAULT_QEMU_MACHINE,
    cores: str = DEFAULT_QEMU_CORES,
    bios: Optional[str] = None,
    detach: bool = False,
    custom: str = "",
    **kwargs,
  ):
    if not which(qemu):
      helper.error(qemu + " isn't installed")

    if packages is None:
      packages = list()

    if not os.path.exists(Qegd.QEMU_DIR):
      helper.info(f"Generating {Qegd.QEMU_DIR} dir")
      os.makedirs(Qegd.QEMU_DIR)

    if arm:
      qemu = Qegd.DEFAULT_QEMU_ARM_CMD if qemu == Qegd.DEFAULT_QEMU_CMD else qemu
      cpu = Qegd.DEFAULT_QEMU_ARM_CPU if cpu == Qegd.DEFAULT_QEMU_CPU else cpu
      machine = Qegd.DEFAULT_QEMU_ARM_MACHINE if machine == Qegd.DEFAULT_QEMU_MACHINE else machine
      bios = Qegd.DEFAULT_QEMU_ARM_BIOS if bios is None else bios

    self._img = img
    self._user = user
    self._forward = forward if forward else dict()
    self._qemu = qemu
    self._cpu = cpu
    self._memory = memory
    self._cores = cores
    self._machine = machine
    self._bios = bios
    self._custom = custom
    self._detach = detach

    self._vm_setup()

    packages += Pwngd.DEFAULT_PACKAGES

    super().__init__(
      binary=binary,
      user=user,
      host=self._host,
      port=self._port,
      packages=packages,
      **kwargs,
    )

  @staticmethod
  def _is_local(url) -> bool:
    """
    check if provided url is local or remote
    :param url: url to check
    :return: if the url is local or remote
    """
    url_parsed = urlparse(url)
    if url_parsed.scheme in ("file", ""):  # Possibly a local file
      return os.path.exists(url_parsed.path)
    return False

  CURRENT_IMG = QEMU_DIR + "current.img"

  def _set_local_img(self):
    """
    get local image for qemu machine
    """
    if Qegd._is_local(self._img):
      helper.info("Using local image")
      self._local_img = self._img
    else:
      if not os.path.exists(Qegd.IMGS_DIR):
        os.makedirs(Qegd.IMGS_DIR)
      self._local_img = Qegd.IMGS_DIR + urlparse(self._img).path.rsplit("/", 1)[-1]
      if not os.path.exists(self._local_img):
        helper.info("online qemu image starting download")
        urlretrieve(self._img, self._local_img)
        helper.info(f"saved image to {self._local_img}")
    copyfile(self._local_img, Qegd.CURRENT_IMG)

  METADATA_FILE = QEMU_DIR + "metadata.yaml"
  _METADATA = """instance-id: iid-local01
local-hostname: cloudimg
"""
  USER_DATA_FILE = QEMU_DIR + "user-data.yaml"
  _USER_DATA = """#cloud-config
users:
  - default
  - name: {user}
    groups: sudo
    shell: /bin/bash
    sudo: ['ALL=(ALL) NOPASSWD:ALL']
    ssh_authorized_keys:
      - {pubkey}
"""

  SEED_FILE = QEMU_DIR + "seed.img"

  _GENERATE_SEED_IMG = f"cloud-localds {SEED_FILE} {USER_DATA_FILE} {METADATA_FILE}"

  def _setup_seed(self):
    """
    create seed.img with config data like ssh keypair in .qemu
    """
    if not which("cloud-localds"):
      helper.error("cloud-image-utils is not installed")

    if os.path.exists(Qegd.SEED_FILE):
      os.remove(Qegd.SEED_FILE)
    with open(Qegd.METADATA_FILE, "w") as metadata_file:
      metadata_file.write(Qegd._METADATA)
    helper.generate_keypair()
    with open(Qegd.USER_DATA_FILE, "w") as user_data_file:
      with open(Pwngd.PUBKEYFILE, "r") as pubkey_file:
        pubkey = pubkey_file.readline()
      user_data_file.write(Qegd._USER_DATA.format(pubkey=pubkey, user=self._user))
    os.system(Qegd._GENERATE_SEED_IMG)

  _QEMU_PORT_FORWARDING = ",hostfwd={type}::{guest}-:{host}"
  _QEMU_START = (
    "{qemu} "
    + "{machine} "
    + "{cores} "
    + "{cpu} "
    + "{memory} "
    + "-nographic "
    + "{bios} "
    + "-device virtio-net-pci,netdev=net0 "
    + "-netdev user,id=net0,hostfwd=tcp::{port}-:22"
    + "{ports} "
    + "-drive if=virtio,format=qcow2,file={img} "
    + "-drive if=virtio,format=raw,file={seed} "
    + "{custom} "
  )

  _QEMU_PIPE = "&> /dev/null; "

  _QEMU_SUFFIX = "rm {lock} {current}"

  _QEMU_ARM_START = ""
  LOCKFILE = QEMU_DIR + "qemu.lock"

  def _qemu_start(self):
    """
    start qemu machine
    """
    self._lock(Qegd.TYPE)
    helper.info(f"starting qemu machine, ssh port {self._port}")
    with open(Qegd.LOCKFILE, "w") as lockfile:
      lockfile.write(str(self._port))
    port_forwarding = "".join(
      Qegd._QEMU_PORT_FORWARDING.format(
        type=host.split("/")[1], host=host.split("/")[0], guest=guest
      )
      for host, guest in self._forward.items()
    )
    qemu_cmd = Qegd._QEMU_START.format(
      qemu=self._qemu,
      machine=f"{Qegd.DEFAULT_QEMU_MACHINE_PREFIX} {self._machine}" if self._machine else "",
      cores=f"{Qegd.DEFAULT_QEMU_CORES_PREFIX} {self._cores}" if self._cores else "",
      cpu=f"{Qegd.DEFAULT_QEMU_CPU_PREFIX} {self._cpu}" if self._cpu else "",
      memory=f"{Qegd.DEFAULT_QEMU_MEMORY_PREFIX} {self._memory}" if self._memory else "",
      bios=f"{Qegd.DEFAULT_QEMU_BIOS_PREFIX} {self._bios}" if self._bios else "",
      port=self._port,
      ports=port_forwarding,
      img=Qegd.CURRENT_IMG,
      custom=self._custom if self._custom else "",
      seed=Qegd.SEED_FILE,
    )
    qemu_suffix = Qegd._QEMU_SUFFIX.format(
      lock=f"{Qegd.LOCKFILE} {Pwngd.LOCKFILE}", current=Qegd.CURRENT_IMG
    )

    helper.info(qemu_cmd)
    if self._detach:
      pwnlib.util.misc.run_in_new_terminal(qemu_cmd, kill_at_exit=False)
    else:
      pid = os.fork()
      if pid == 0:
        if self._detach:
          pass
        else:
          qemu_cmd += Qegd._QEMU_PIPE + qemu_suffix
          os.system(qemu_cmd)
        exit(0)

  def _new_vm(self) -> None:
    """
    create new vm
    """
    self.is_new = True

    helper.info(f"no Lockfile in {Qegd.LOCKFILE}")
    self._set_local_img()
    self._setup_seed()
    self._port = helper.first_free_port(Qegd.DEFAULT_PORT)
    # start qemu in independent process
    self._qemu_start()
    time.sleep(15)

  def _vm_setup(self) -> None:
    """
    setup qemu machine
    """
    self._host = Qegd.DEFAULT_HOST
    if not os.path.exists(Qegd.LOCKFILE):
      self._new_vm()
    else:
      with open(Qegd.LOCKFILE, "r") as lockfile:
        self._port = int(lockfile.readline())
      if not helper.is_port_in_use(self._port):
        helper.info(
          f"Lockfile in {Qegd.LOCKFILE}, but no running machine detected. Creating new one"
        )
        os.remove(Qegd.LOCKFILE)
        self._new_vm()
      helper.info(f"Lockfile in {Qegd.LOCKFILE}. Using running qemu instance at port {self._port}")
